import logging
from itsdangerous import BadSignature
from tornado import ioloop
import datetime


logger = logging.getLogger('tornadopush')


class ChannelMixin(object):
    @property
    def channel(self):
        return getattr(self, '_channel', None)

    @channel.setter
    def channel(self, channel):
        if not channel:
            channel = self.application.settings['default_channel']
        allowed_channels = self.application.settings['allowed_channels']
        if allowed_channels and channel not in allowed_channels:
            self.send_error(404)
        self._channel = channel

    @property
    def channel_key(self):
        return self.application.settings['channel_format'] % self.channel


class AuthentifyMixin(object):
    @property
    def auth_enabled(self):
        return self.application.settings['auth_enabled']

    def authentify(self, token=None, callback=None):
        self.user_id = None
        if not callback:
            callback = self.on_authorized
        if token:
            self.user_id = self._validate_token(token)
        if self.user_id is None and self.auth_enabled:
            logger.debug('Token %s rejected' % token)
            self.on_unauthorized()
        else:
            if self.user_id:
                logger.debug('User %s connected on #%s' % (self.user_id, self.channel))
            else:
                logger.debug('New anonymous connected on #%s' % self.channel)
            callback()

    def _validate_token(self, token):
        max_age = self.application.settings['token_max_age']
        try:
            user_id, user_channels = self.application.token_serializer.loads(token, max_age=max_age)
            if not user_channels or self.channel in user_channels:
                return user_id
        except BadSignature as e:
            pass

    def on_unauthorized(self):
        self.send_error(403)

    def on_authorized(self):
        pass


class BrukvaMixin(object):
    def create_client(self):
        self.client = self.application.create_brukva_client()
        self.subscribed = False

    def subscribe(self, *channels):
        def on_subscribe(*args):
            self.client.listen(self.on_redis_message)
            self.subscribed = True
            self.subscribed_to = channels
        self.client.subscribe(channels, on_subscribe)

    def unsubscribe(self, callback=None):
        if not self.client or not self.subscribed:
            if callback:
                callback()
            return
        # a bug in brukva where the unsubscribe callback
        # is called too early forces us to use a timer
        # to check when the operation is finished
        self.client.unsubscribe(self.subscribed_to)
        def check():
            if self.client.subscribed:
                ioloop.IOLoop.instance().add_timeout(datetime.timedelta(0.00001), check)
            else:
                self.subscribed = False
                if callback:
                    callback()
        ioloop.IOLoop.instance().add_timeout(datetime.timedelta(0.00001), check)

    def close_client(self, callback=None):
        def on_unsubscribe(*args):
            if self.client:
                self.client.disconnect()
            logger.debug('Client disconnected')
            if callback:
                callback()
        if self.subscribed:
            self.unsubscribe(on_unsubscribe)
        else:
            on_unsubscribe()


class WebSocketMixin(object):
    def send_error(self, code, **kwargs):
        self.close(code)

    def check_origin(self, origin):
        return True


from tornado.websocket import WebSocketHandler
from .base import *


class PresenceMixin(object):
    @property
    def presence_users_key(self):
        return self.application.settings['presence_users_key'] % self.channel

    @property
    def presence_channel(self):
        return self.application.settings['presence_channel'] % self.channel

    def notify_join(self, callback=None):
        def on_members(members):
            if members:
                self.handle_presence("+" + ";+".join(members))
            if callback:
                callback()
        def list_members(*args):
            self.client.smembers(self.presence_users_key, on_members)
        def on_add(r):
            logger.debug('Presence: %s joined' % self.user_id)
            self.client.publish(self.presence_channel, "+%s" % self.user_id,
                list_members)
        if self.user_id:
            self.client.sadd(self.presence_users_key, self.user_id, callback=on_add)
        else:
            list_members()

    def notify_leave(self, callback=None):
        def on_published(r):
            logger.debug('Presence: %s left' % self.user_id)
            if callback:
                callback()
        def on_removed(r):
            self.client.publish(self.presence_channel,
                "-%s" % self.user_id, on_published)
        self.client.srem(self.presence_users_key, self.user_id, callback=on_removed)

    def handle_presence(self, msg):
        raise NotImplementedError


class PresenceHandler(PresenceMixin, ChannelMixin, AuthentifyMixin, BrukvaMixin,\
                      WebSocketMixin, WebSocketHandler):

    def __init__(self, *args, **kwargs):
        WebSocketHandler.__init__(self, *args, **kwargs)
        self.create_client()

    def open(self, channel=""):
        self.channel = channel
        self.authentify(self.get_argument('token', None))

    def on_authorize(self):
        self.notify_join(lambda: self.subscribe(self.presence_channel))

    def on_message(self, message):
        pass

    def on_redis_message(self, msg):
        self.handle_presence(str(msg.body))

    def handle_presence(self, msg):
        self.write_message("presence:%s" % msg)

    def on_close(self):
        def on_unsubscribe():
            self.notify_leave(lambda: self.close_client())
        self.unsubscribe(on_unsubscribe)